package IrssiFetcher;

use strict;
use warnings;
use Errno qw(EINPROGRESS);
use Fcntl;
use Irssi;
use POSIX qw/BUFSIZ/;
use Socket;

use vars qw/$DNSCache @ISA @EXPORT @EXPORT_OK/;
require Exporter;
@ISA = qw(Exporter AutoLoader);
@EXPORT = qw();
@EXPORT_OK = qw(fetch);

# Local variables.
my %_tags;
my %_dns;
my $caller;

=head1 NAME

 IrssiFetcher - nonblocking http fetcher

=head1 SYNOPSIS

 use IrssiFetcher qw/fetch/;
 fetch("http://www.google.com/", \&googleFetched);

 sub googleFetched
 {
  my ($content, $headers, $is_success) = @_;
  unless($is_success)
  {
   Irssi::print("Noooes, fetching of googlef failed!");
   return;
  }
  
  Irssi::print("Yeah, google fetched, it looks like this:".$content);
 }

=head1 DESCRIPTION

Since Irssi doesn't give it to us directly we'll have to use some tricks to be able to
fetch an url without having Irssi stalled for a few seconds.
This library allows plugins to fetch a URL without blocking Irssi.

Since DNS resolution is done by irssi that might still stall your IRC client. In order to prevent that can cache
dns requests for one hour. To change this set the irssi setting 'irssifetcher_dnscache' to the amounts of seconds it may be cached.
Defaults to 5.

It took me some fancy trickery to get this library to work in Irssi, because Irssi doesn't like it when a script loads
modules that use Irssi themselves. Irssi requires a script to blame for when stuff goes wrong I guess, which fails when
we set callbacks directly (like Irssi::input_add etc). Therefore on the first call to fetch we hack a wrapper function into the 
caller's namespace which we use for the Irssi functions.

=head1 METHODS
 
=over
 
=item fetch ( $url, $callback, [ $cbarg,  $options ] )

Fetches $url and calls the callback when done.

cbarg is optional al will be passed back to the callback.
Callback should look like: 
	sub cbFetch { my ($content, $headers, $success, $cbarg) = @_; <some useful stuff here> }
Where:
	content: a >>>REFERENCE<<< to the result scalar that comes after the header. Note that this content could be an empty string, but never undef. Did I mention REFERENCE?! ;)
	headers: a REFERENCE to the headers hash with result headers, such as "Content-Length" => 123
	success: 1 on success 0 otherwise. Currently a [200] is considered success, the rest will either be handled or is a failure.

Options is an optional hash reference with:
 postdata: makes it a POST request instead of GET with this data.
 headers : hash with extra request headers, such as { "X-Foo" => "Bar" }

Returns 0 on success, error string on failure.

NOTE: returns failure on something it's already fetching. This is a feature. Maybe an option later ;)

=cut

sub fetch
{
	my ($url, $cb, $cbargs, $opt) = @_;

	_init(caller()) unless(defined($caller));

	# Check if already running
	if (defined($_tags{$url}))
	{
		if (defined($opt) && ref($opt) eq 'HASH' && $opt->{'restart'})
		{
			_debug(1, "Fetch for $url already running - restarting!");
			_irssi("Irssi::input_remove", $_tags{$url}{'tag'}) if(defined($_tags{$url}{'tag'}));
			delete $_tags{$url};
		}
		else
		{
			return "Error: already fetching $url";
		}
	}

	$_tags{$url}{'cb'}      = $cb;
	$_tags{$url}{'opt'}     = $opt;
	$_tags{$url}{'cbarg'}   = $cbargs;
	$_tags{$url}{'content'} = '';
	$_tags{$url}{'headers'} = {};
	$_tags{$url}{'offset'}  = 0;
	$_tags{$url}{'reqpath'} = '';
	$_tags{$url}{'success'} = 0;
	$_tags{$url}{'time_e'}  = 0;
	$_tags{$url}{'time_s'}  = time;

	my $flags;
	my ($host, $port);
	my $tmp;
	my $sock;
	unless ($url =~ m/http:\/\/(:(\d+))?([^\/]+)(\/.*)$/)
	{
		my $error = "Could not parse $url, typo?";
		_finisher($url);
		return $error;
	}
	$host = $3;
	$port = 80;
	$port = $2 if (defined($2));
	$_tags{$url}{'reqpath'} = $4;

	# First let's resolve the host.
	my $dnscache = _irssi("Irssi::settings_get_int", 'irssifetcher_dnscache');
        $tmp = undef;
        if ($dnscache && defined($_dns{$host}))
        {
		if (time < $_dns{$host}{'expire'})
		{
			_debug(1, "DNS Cache HIT");
			$tmp = $_dns{$host}{'cache'};
		}
		else
		{
			_debug(1, "DNS Cache expired hit -> MISS");
			delete $_dns{$host};
		}
        }
	else
	{
		_debug(1, "DNS Cache MISS");
	}
        unless (defined($tmp))
        {
		# TODO: make this nonblocking
                $tmp = inet_aton($host);
        }
        unless ($tmp)
        {
                my $error = "Error resolving $host: $!";
                _finisher($url);
		return $error;
        }
	if ($dnscache)
	{
		$_dns{$host}{'cache'} = $tmp;
		$_dns{$host}{'expire'} = time + $dnscache;
	}

	if (!socket($sock, PF_INET, SOCK_STREAM, getprotobyname('tcp')))
	{
		my $error = "Error creating socket: $!";
		_finisher($url);
		return $error;
	}
	$_tags{$url}{'sock'} = $sock;

	if (!defined($flags = fcntl($sock, F_GETFL, 0)))
	{
		my $error = "Error getting socket flags: ".$!;
		_finisher($url);
		return $error;
	}
	if (!defined(fcntl($sock, F_SETFL, $flags | O_NONBLOCK)))
	{
		my $error = "Error setting non blocking: ".$!;
		_finisher($url);
		return $error;
	}

	# Now that we have a nonblocking socket, make it connect to the target.
	$tmp = sockaddr_in($port,$tmp);                                                                                                                                        
	unless(connect($sock, $tmp))
	{
		if ($! == EINPROGRESS)
		{
			# No problem here, it's working.
		}
		else
		{
			my $error = "Error connecting to host: $!";
			_finisher($url);
			return $error;
		}
	}

	$_tags{$url}{'writebuf'} = _getRequestHeaders($url);
	_debug(1, "Write buffer set: ".$_tags{$url}{'writebuf'});
	$_tags{$url}{'tag'} = _irssi("Irssi::input_add", fileno($sock), INPUT_WRITE, \&_cbConnected, $url);
	return 0;
}

sub _init
{
	my ($c) = @_;

	$caller = $c;

	local $SIG{'__DIE__'};
	eval ("{ package $caller;".'sub irssiWrapper { no strict "refs"; my $func = shift; return &{$func}(@_); } }');
	# Irssi settings
	_irssi("Irssi::settings_add_int", 'irssifetcher', 'irssifetcher_dnscache', 5); # DNS Cache, 0 is disabled.
	_irssi("Irssi::settings_add_int", 'irssifetcher', 'irssifetcher_debug', 0); # Debuglevel, 0 only prints errors.
}

sub _irssi
{
	no strict 'refs';
	# Maybe stuff got reloaded... like the parent script, which fails to reload this library.
	unless(defined($caller))
	{
		# This shouldn't happen unless you have a feeling for terrible timing or multiple scripts using this library while unloading the first caller orso... whatever.
		_debug("Uh oh, this breaks. Type /upgrade and reload the script that causes this. Caller stack: ".join(", ", caller(), caller(1), caller(2)));
		return undef;
	}
	unless(defined( &{$caller."::irssiWrapper"} ))
	{
		_init($caller);
	}
	return &{$caller."::irssiWrapper"}(@_);
}

sub _getRequestHeaders
{
	my ($url) = @_;
	my $post = 0;
	my $h = "GET ".$_tags{$url}{'reqpath'}." HTTP/1.0\r\n";
	my %headers = (
		'Connection' => 'close',
		'User-Agent' => 'IrssiFetcher v0.1'
	);
	if (defined($_tags{$url}{'opt'}) && ref($_tags{$url}{'opt'}) eq 'HASH')
	{
		if (defined($_tags{$url}{'opt'}->{'postdata'}))
		{
			$h = "POST ".$_tags{$url}{'regpath'}." HTTP/1.0\r\n";
			$headers{'Content-Length'} = length($_tags{$url}{'opt'}->{'postdata'});
			$headers{'Content-Type'} = "application/x-www-form-urlencoded";
			$post = 1;
		}
	}
	foreach my $h (keys %headers)
	{
		$h .= "$h: $headers{$h}\r\n";
	}
	$h .= "\r\n";
	$h .= $_tags{$url}{'opt'}->{'postdata'} if ($post);
	return $h;
}

# Connection established. Write request to server.
sub _cbConnected
{
	my ($url) = @_;
	_irssi("Irssi::input_remove", $_tags{$url}{'tag'});
	my $sock = $_tags{$url}{'sock'};
        my $res = syswrite($sock, $_tags{$url}{'writebuf'}, POSIX::BUFSIZ);
        unless(defined($res))
        {
                _debug("Error writing to socket: ".$!." for url $url");
		_finisher($url);
                return undef;
        }
        _debug(1, "Wrote $res bytes, cutting those from write buffer");
        substr($_tags{$url}{'writebuf'}, 0, $res) = '';

	# If we aren't done writing yet, set another callback. Otherwise set the read callback.
	if (length($_tags{$url}{'writebuf'}))
	{
		$_tags{$url}{'tag'} = _irssi("Irssi::input_add", fileno($sock), INPUT_WRITE, \&_cbConnected, $url);
	}
	else
	{
		$_tags{$url}{'tag'} = _irssi("Irssi::input_add", fileno($sock), INPUT_READ, \&_cbReader, $url);
	}
}

sub _cbReader
{
	my ($url) = @_;
	_irssi("Irssi::input_remove", $_tags{$url}{'tag'});
	my $sock = $_tags{$url}{'sock'};
	my $res;
	if (defined($res = sysread($sock, $_tags{$url}{'content'}, POSIX::BUFSIZ, $_tags{$url}{'offset'})))                                                                 
	{
		_debug(1, "IrssiFetcher data read, $res, offset $_tags{$url}{'offset'}");
		$_tags{$url}{'offset'} += $res;
		if ($res == 0)
		{
			_debug(1, "IrssiFetcher Done reading for url [$url]...");
			_parseResult($url);
			close $_tags{$url}{'sock'};
			return _finisher($url);
		}
	}
	$_tags{$url}{'tag'} = _irssi("Irssi::input_add", fileno($sock), INPUT_READ, \&_cbReader, $url);
}

sub _parseResult
{
	my ($url) = @_;
	# Split header from content and see if this was a success.
	unless ($_tags{$url}{'content'} =~ /^(.*?)\r\n\r\n(.*)$/s)
	{
		$_tags{$url}{'success'} = 0;
		_debug("Weird result received from fetch.... no headers, maybe not a webserver we contacted? :-p. Whatever, we're done.");
		return;
	}
	my $headers = $1;
	$_tags{$url}{'content'} = $2;

	# First line is the result line, should be something like 'HTTP/1.1 200 OK\r\n'
	if ($headers =~ s/^HTTP\/\d+\.\d+\s+(\d+)(\s+.*?)\r\n//s)
	{
		my $code = $1;
		my $status = $2;
		$_tags{$url}{'success'} = ($code >= 200 && $code < 300?1:0);
		$_tags{$url}{'success_code'} = $code;
		$_tags{$url}{'success_status'} = $status;
	}
	foreach my $h (split /\r\n/, $headers)
	{
		if ($h =~ /^([^:]+):(.*)$/)
		{
			$_tags{$url}{'headers'}{$1} = $2;
		}
		else
		{
			_debug("Weird header: [$h] - ignoring!");
		}
	}

	# TODO: handle redirects here. (simply start another fetch).
	if ($_tags{$url}{'success_code'} >= 300 && $_tags{$url}{'success_code'} < 400)
	{
		_debug("Looks like a redirect but we don't handle those yet. Sorry ;-p");
		return;
	}
}

sub _debug
{
	my ($level, @msg) = @_;
	my $debuglevel = _irssi("Irssi::settings_get_int", 'irssifetcher_debug');
	unless(scalar(@msg))
	{
		@msg = ($level);
		$level = 0;
	}
	return if($level > $debuglevel);
        my $flags = MSGLEVEL_CLIENTNOTICE;
        $flags |= MSGLEVEL_NO_ACT if ($level > 1);

	foreach(@msg)
	{
		Irssi::print("[\cC07IrssiFetcher.pm\co] ".$_,$flags);
	}
        return 1;
}

# Internal function, calls callback and cleans up state info for a certain url.
sub _finisher
{
	my ($url) = @_;
	local $SIG{'__DIE__'};
	$_tags{$url}{'time_e'} = time;
	eval
	{
		if ($_tags{$url}{'cbarg'})
		{
			$_tags{$url}{'cb'}(\$_tags{$url}{'content'}, $_tags{$url}{'headers'}, $_tags{$url}{'success'}, $_tags{$url}{'cbarg'});
		}
		else
		{
			$_tags{$url}{'cb'}(\$_tags{$url}{'content'}, $_tags{$url}{'headers'}, $_tags{$url}{'success'});
		}
	};
	delete $_tags{$url};
}

=back
 
=head1 AUTHOR
  
  BenV / JuNe <benv-necta@junerules.com>
   
=cut

1;
