use strict;
use warnings;
use Irssi qw/INPUT_READ INPUT_WRITE input_add settings_get_bool settings_add_bool settings_add_str settings_get_str settings_set_str settings_add_int settings_get_int/;
use Irssi::TextUI;
use vars qw/$VERSION %IRSSI/;
use lib::again;

=head1 NAME

 necta - Nectarine / Scenemusic.net Irssi Plugin

=head1 SYNOPSIS

 Put this script in  your Irssi scripts dir: ~/.irssi/scripts (create it if it doesn't exist) and the Nectarine.pm and IrssiFetcher.pm in your scripts/lib/ dir.
 Then type /script load necta

 For automatic loading link the script in the ~/.irssi/scripts/autorun directory (create if it's missing) like this:
 cd ~/.irssi/scripts/autorun ; ln -s ../necta.pl

 To create a statusbar with currently playing and upcoming song info, load the script and type:
 /statusbar necta ENABLE
 /statusbar necta PLACEMENT top
 /statusbar necta POSITION 1
 /statusbar necta ADD -alignment left necta

=head1 DESCRIPTION

This plugin uses lib/Nectarine.pm with our own lib/Irssi/Fetcher.pm to fetch current stream information
and show this in a irssi statusbar.

Note: I tried to support the /format joke, but due to a lack of documentation on the Irssi part I gave up on it
after fucking around with $theme->format_expand for more than a day and getting absolutely no progress.
If you feel like supporting it, please feel free to send me a patch.

=head1 AUTHORS

BenV / JuNe <benv-necta@junerules.com>

=cut

%IRSSI = (
	authors         => 'BenV / JuNe',
	contact         => 'benv-necta@junerules.com',
	name            => 'necta',
	description     => 'Provide realtime nectarine stream information',
	license         => 'BSD',
	modules         => 'IrssiFetcher, Nectarine',
	url             => 'http://notes.benv.junerules.com/necta/',
	changed         => '11-04-2010'
);
$VERSION = "0.02";

push @INC, Irssi::get_irssi_dir()."/scripts/lib/";
init() || die "Load error. Make sure you also put the required perl modules (IrssiFetcher and Nectarine) in the right place!";

my $necta;
my $refresh_tag;  # Statusbar refresh timer tag
my $refresh_time; # Keeps track of statusbar refresh
my $update_tag;   # Nectarine queue updater timer tag
my $update_fail_timer; # Amount of seconds to wait upon failure. Doubles every time it fails until timer > 600, it will stop trying with a warning.

sub init
{
	initSettings();
	$update_fail_timer = 30;
	Irssi::command_bind('necta', \&cmd_necta);
	loadModule("IrssiFetcher", "fetch") || return 0;
	loadModule("Nectarine") || return 0;
	my $debuglevel = settings_get_int('necta_debug');
	$necta = Nectarine->new({ debug => \&debug, fetcher => \&nectaFetch, verbose => $debuglevel });
	necta_updater();
	Irssi::statusbar_item_register('necta', 0, 'nectabar');
	Irssi::statusbars_recreate_items();
	$refresh_tag = Irssi::timeout_add(1000, 'refresh_nectabar', undef);
	Irssi::signal_add('complete word', \&tabCompletion);
	# Possibly for later version:
	# Irssi::signal_add_last('setup changed', \&redraw);
	printHelp();
	return 1;
}

# Wrapper function to make Irssi silent about namespace issues....
sub nectaFetch
{
	return fetch(@_);
}

sub cmd_necta
{
	my ($data, $server, $witem) = @_;

	if ($data =~ /^dump$/)
	{
		use Data::Dumper;
		debug(Dumper($necta));
		return;
	}
	elsif ($data =~ /^update$/)
	{
		debug("Manually requesting necta queue update. Please stand by.");
		necta_updater(1);
		return;
	}
	elsif ($data =~ /^queue\s*(\d+)?$/)
	{
		my $amount = $1 // 3;
		my $r = necta_queue($amount);
		unless(defined($r))
		{
			debug("No queue information present!");
			return;
		}
		map { debug($_) } "Necta queue information. Upcoming $amount entries:", @$r;
		return;
	}
	printHelp();
	return;
}

sub printHelp
{
	my @help = (
		"{g}Nectarine / Scenemusic.eu{/}! Version {c}$VERSION\{/}.",
		 "To create a statusbar with currently playing and upcoming song info, load the script and type:",
		 "/statusbar necta ENABLE",
		 "/statusbar necta PLACEMENT top",
		 "/statusbar necta POSITION 1",
		 "/statusbar necta ADD -alignment left necta",
		 "",
		 "To change the layout of the bars, check out {y}/set necta{/} and change what you want",
		 "Also checkout the other settings there (like disabling colors).",
		 "",
		 "{y}Avaible commands:{/}",
		 " {c}*{/} {Gr}/necta queue [1-9]{/} - shows current [1-9] queue entries",
		 " {c}*{/} {Gr}/necta update     {/} - starts a queue update. Is done automatically by script, but you can force an update using this.",
	);
	foreach(@help)
	{
		debug($_);
	}
}

sub tabCompletion
{
	my ($cl, $win, $word, $ls, $ws) = @_;

	my %cmds = (
		help   => [],
		queue  => [],
		update => [],
		);
	my @opts;
	return unless ($ls =~ s#^\s*/necta\s*##i); # Strip any command prefix
	debug("Tab completion debug (after strip): word[$word] ls[$ls] win[$win]",3);
	if ($ls eq '')
	{
		if ($word eq '')
		{
			@opts = keys %cmds;
		}
		else
		{
			@opts = grep(/^$word/, keys %cmds);
		}
		debug("Adding to completion list: [".join(",",@opts)."]",3);
		push @$cl, @opts;
	}
	else
	{
		return unless defined($cmds{$ls});
		debug("Searching specific commands for completion list... [$word]",3);
		if ($word eq '')
		{
			@opts = @{$cmds{$ls}};
		}
		else
		{
			@opts = grep {/^$word/} @{$cmds{$ls}};
		}
		debug("Adding to completion list: [".join(",",@opts)."]",3);
		push @$cl, @opts if ($#opts >= 0);
	}
	return;
}

sub necta_queue
{
	# Returns last X queue entries (or less if less info is present) in readable format.
	my ($amount) = @_;
	my @r;
	unless (exists($necta->{'queue'}[0]) && scalar(keys(%{$necta->{'queue'}[0]})))
	{
		return undef;
	}

	my $f_queue = settings_get_str('necta_queue');
	for (my $i=1; $i <= $amount; $i++)
	{
		return \@r unless(exists($necta->{'queue'}[$i - 1]));
		my $row = $f_queue;
		foreach my $k (keys %{$necta->{'queue'}[$i - 1]}, 'pos')
		{
			my $tmp = $necta->{'queue'}[$i - 1]{$k} // "";
			$tmp = $i if ($k eq 'pos');
			$row =~ s/%$k%/$tmp/g;
		}
		push @r, $row;
	}
	return \@r;
}

sub initSettings
{
	settings_add_int('necta','necta_debug', 0); # Debug level
	settings_add_bool('necta','necta_color', 1); # Quick enable/disable switch for colors in status messages.
	settings_add_int('necta','necta_refresh', 1); # Amount of seconds between statusbar updates. Set higher if you're on a slow link.
	# Theming
	settings_add_str('necta','necta_bar', '%now% %next% %queue%'); # Global bar layout
	settings_add_str('necta','necta_bar_now', '{sb %YNecta%c: %G%artist% %c- %P%song% %c(%W%timeleft%%c/%r%song_length%%c)}');
	settings_add_str('necta','necta_bar_next', '{sb %YNext%c: %G%artist% %c- %P%song%}');
	settings_add_str('necta','necta_bar_queue', '{sb %YPending%c: %G%pending%}');
	# for /necta queue
	settings_add_str('necta','necta_queue', '%W%pos%. %G%artist% %c- %P%song% %c(%song_length%%c)');
}

sub loadModule
{
	my ($module, $imports) = @_;
	my $loadstring = "use again '" . $module . "'";
	if (defined($imports) && length($imports))
	{
		$loadstring .= ", [qw/$imports/]";
	}
	local $SIG{'__DIE__'};
	eval ($loadstring);
	return 1;
}

sub debug
{
	my ($msg, $level) = @_;
	$level = 0 unless(defined($level));
	my $debuglevel = settings_get_int('necta_debug');
	return 1 if ($level > $debuglevel);

	my $flags = MSGLEVEL_CLIENTNOTICE;
        $flags |= MSGLEVEL_NO_ACT if ($level > 1);

	$msg = "[{g}necta{/}] ".$msg;
	Irssi::print(colorize_string($msg),$flags);
	return 1;
}

sub colorize_string
{
	my ($msg) = @_;

	# Color codes.
        my %c = (
                b  => "\cC02", # blue
                g  => "\cC03", # green
                R  => "\cC04", # light red
                r  => "\cC05", # red
                p  => "\cC06", # purple
                o  => "\cC07", # orange
                y  => "\cC08", # yellow
                G  => "\cC09", # light green
                c  => "\cC10", # cyan
                C  => "\cC11", # light cyan
                B  => "\cC12", # light blue
                P  => "\cC13", # light purple
                gr => "\cC14", # gray
                Gr => "\cC15", # light gray
                n  => "\co"  , # no effects
                '/'  => "\co"  , # no effects
                bo => "\cb"  , # bold
                ul => "\c_"    # underline
        );

	my $color = settings_get_bool('necta_color');
	foreach my $cc (keys %c)
	{
		if ($color)
		{
			$msg =~ s/{$cc}/$c{$cc}/gx;
		}
		else
		{
			$msg =~ s/{$cc}//g;
		}
	}
	return $msg.$c{'/'};
}

# Bars
sub nectabar
{
	my ($item, $get_size_only) = @_;
	my $format = settings_get_str('necta_bar');
	my $f_now = settings_get_str('necta_bar_now');
	my $f_next = settings_get_str('necta_bar_next');
	my $f_queue = settings_get_str('necta_bar_queue');

	if (keys(%{$necta->{'now'}}))
	{
		$necta->updateTimeleft();
		foreach my $k (keys %{$necta->{'now'}})
		{
			my $tmp = $necta->{'now'}{$k} // "";
			$f_now =~ s/%$k%/$tmp/g;
		}
	}
	else
	{
		foreach my $k (qw/artist artist_id song song_id song_length timeleft/)
		{
			$f_now =~ s/%$k%/?/g;
		}
	}

	if (defined($necta->{'queue'}[0]) && scalar(keys(%{$necta->{'queue'}[0]})))
	{
		foreach my $k (keys %{$necta->{'queue'}[0]})
		{
			my $tmp = $necta->{'queue'}[0]{$k} // "";
			$f_next =~ s/%$k%/$tmp/g;
		}
	}
	else
	{
		foreach my $k (qw/artist artist_id song song_id song_length/)
		{
			$f_next =~ s/%$k%/?/g;
		}
	}

	my $pending = (keys %{$necta->{'queue'}[0]}?scalar(@{$necta->{'queue'}}):0);
	$f_queue =~ s/%pending%/$pending/g; #

	$format =~ s/%now%/$f_now/g;
	$format =~ s/%next%/$f_next/g;
	$format =~ s/%queue%/$f_queue/g;

	# Check if color is disabled
	unless (settings_get_bool('necta_color'))
	{
		# Great, get rid of all colorcodes in the bars.
		$format =~ s/%[krgybmpcwKRGYBMPCW]//g;
	}

	$item->default_handler($get_size_only, $format, 0, 0);
}

sub refresh_nectabar
{
        my $refresh = settings_get_int('necta_refresh');
        if (!defined($refresh_time) || $refresh_time ne $refresh)
        {
                Irssi::timeout_remove($refresh_tag) if (defined($refresh_tag));
                $refresh = ($refresh>0?$refresh:5); # Require at least 1 second delay.
                $refresh_time = $refresh;
                $refresh_tag = Irssi::timeout_add($refresh*1000, 'refresh_nectabar', undef);
        }
	Irssi::statusbar_items_redraw('necta');
}

# Updater, updates nectarine queue info and schedules next update request.
sub necta_updater
{
	my ($force) = @_;
	$force //= 0;
	Irssi::timeout_remove($update_tag) if ($update_tag); # necta_update() will fix it.
	$update_tag = undef;
	debug("Starting necta update.", 2);
	unless ($necta->updateQueue($force, \&necta_updated))
	{
		debug("Uh oh, update could not be scheduled. Maybe it's still running? Scheduling forced retry in {c}$update_fail_timer\{/} seconds!");
		$update_tag = Irssi::timeout_add(1000 * $update_fail_timer, 'necta_updater', 'force');
		$update_fail_timer *= 2;
		return;
	}
}

sub necta_updated
{
	Irssi::timeout_remove($update_tag) if ($update_tag);
	my ($error) = @_;
	if (defined($error) && $error)
	{
		debug("{r}Failure{/} updating Necta queue info... error: {y}$error\{/}.", 1);
		if ($update_fail_timer >= 600)
		{
			debug("{R}Warning{/} This was the last automatic retry failure. In order to restart the necta queue update, type: {c}/necta update{/}.", 0);
			debug("This means that {r}necta won't update queue information{/} until the above command is typed or the script is reloaded!", 0);
			return;
		}
		else
		{
			debug("{r}Warning{/} Rescheduling update in $update_fail_timer seconds", 1);
		}
		$update_tag = Irssi::timeout_add(1000 * $update_fail_timer, 'necta_updater', undef);
		$update_fail_timer *= 2;
		return;
	}
	$update_fail_timer = 30;
	# Get seconds until next song
	unless(defined($necta->{'now'}) && defined($necta->{'now'}{'timeleft_s'}))
	{
		debug("Hmm, no timeleft information present. Setting queue updater to 4 minutes.", 0);
		$update_tag = Irssi::timeout_add(60 * 4 * 1000, 'necta_updater', undef);
		return;
	}
	my $timeleft = $necta->{'now'}{'timeleft_s'};
	if ($timeleft < 15)
	{
		debug("Increasing next update timer to 15 seconds instead of {c}$timeleft\{/} to prevent hammering.", 2);
		$timeleft = 15;
	}
	debug("Necta update completed {g}successfully{/}. Next update in {c}${timeleft}{/} seconds.", 2);
	$update_tag = Irssi::timeout_add(1000 * $timeleft, 'necta_updater', undef);
}
